from pweb_form_rest.common.pweb_com_fr_message import ComFRMessage
from pweb_form_rest.crud.pweb_crud_common import PWebCRUDCommon
from pweb_form_rest.schema.pweb_rest_schema import PWebDataDTO
from pweb_form_rest.crud.pweb_file_data_crud import FileDataCRUD
from pweb_orm import PWebBaseModel


class RESTDataCRUD(PWebCRUDCommon):
    file_data_crud = None

    def __init__(self, model: PWebBaseModel):
        self.model = model
        self.file_data_crud = FileDataCRUD(model=model)

    def update_and_get_model(self, request_dto: PWebDataDTO, data: dict = None, existing_model=None, query=None):
        if not data:
            data = self.get_json_data(request_dto)
        return self.edit(model_id=data['id'], data=data, request_dto=request_dto, existing_model=existing_model, query=query)

    def create_and_get_model(self, request_dto: PWebDataDTO, data: dict = None):
        if not data:
            data = self.get_json_data(request_dto)
        return self.save(data=data, request_dto=request_dto)

    def create(self, request_dto: PWebDataDTO, response_dto: PWebDataDTO = None, response_message: str = ComFRMessage.create_success, data: dict = None):
        model = self.create_and_get_model(data=data, request_dto=request_dto)
        return self.message_or_data_response(model, response_dto, response_message)

    def details(self, model_id: int, response_dto: PWebDataDTO, query=None):
        existing_model = self.get_by_id(model_id, exception=True, query=query)
        return self.response_maker.data_response(existing_model, response_dto)

    def update(self, request_dto: PWebDataDTO, response_dto: PWebDataDTO = None, response_message: str = ComFRMessage.update_success, existing_model=None, data: dict = None, query=None):
        model = self.update_and_get_model(request_dto=request_dto, existing_model=existing_model, query=query, data=data)
        return self.message_or_data_response(model, response_dto, response_message)

    def delete(self, model_id: int, response_message: str = ComFRMessage.delete_success, query=None):
        self.soft_remove(model_id=model_id, query=query)
        return self.response_maker.success_message(response_message)

    def hard_delete(self, model_id: int, response_message: str = ComFRMessage.delete_success, query=None):
        self.remove(model_id, query=query)
        return self.response_maker.success_message(response_message)

    def paginated_list(self, response_dto: PWebDataDTO, query=None, search_fields: list = None, sort_field=None, sort_order=None, item_per_page=None):
        data_list = self.read_all(query=query, search_fields=search_fields, sort_field=sort_field, sort_order=sort_order, item_per_page=item_per_page)
        return self.response_maker.paginate_response(data_list, response_dto)

    def list(self, response_dto: PWebDataDTO, query=None, search_fields: list = None, sort_field=None, sort_order=None):
        data_list = self.read_all(query=query, search_fields=search_fields, sort_field=sort_field, sort_order=sort_order, enable_pagination=False)
        return self.response_maker.list_data_type_response(data_list, response_dto=response_dto)

    def upload_file_data(self, request_dto: PWebDataDTO, upload_path, response_dto: PWebDataDTO = None, override_names: dict = None, response_message: str = ComFRMessage.create_success, form_data=None):
        return self.file_data_crud.upload_file_data(request_dto=request_dto, upload_path=upload_path, response_dto=response_dto, override_names=override_names, response_message=response_message, form_data=form_data)

    def update_upload_file_data(self, request_dto: PWebDataDTO, upload_path, response_dto: PWebDataDTO = None, override_names: dict = None, response_message: str = ComFRMessage.update_success, existing_model=None, form_data: dict = None, query=None):
        return self.file_data_crud.update_upload_file_data(request_dto=request_dto, upload_path=upload_path, response_dto=response_dto, override_names=override_names, response_message=response_message, existing_model=existing_model, form_data=form_data, query=query)



from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin
from apispec.ext.marshmallow.field_converter import FieldConverterMixin
import pweb_form_rest.common.pweb_custom_field
from pweb_form_rest.common.pweb_fr_config import PWebFRConfig
from pweb_form_rest.common.pweb_fr_constant import SwaggerCommon, SwaggerDataType, SwaggerDefinitionType
from pweb_form_rest.swagger.pweb_swagger_param_def import SwaggerParamDef
from pweb_form_rest.swagger.pweb_swagger_schema import PWebSwaggerSchema


class PWebSwaggerGenerator:
    _swagger_api_spec: APISpec = None

    def __init__(self):
        self.init_api_spec()

    def init_api_spec(self):
        FieldConverterMixin.field_mapping[pweb_form_rest.common.pweb_custom_field.FileField] = ("string", "binary")
        self._swagger_api_spec = APISpec(
            title=PWebFRConfig.SWAGGER_TITLE,
            version=PWebFRConfig.SWAGGER_APP_VERSION,
            openapi_version="3.0.2",
            plugins=[MarshmallowPlugin()]
        )
        self._init_pweb_schema()

    def process_action_definitions(self, action_definitions: list):
        for definition in action_definitions:
            self.process(definition)

    def process(self, definition: SwaggerParamDef):
        definition.init_schema_key()
        self._entry_spec(definition)

    def get_swagger_spec(self):
        specification = {}
        if self._swagger_api_spec:
            specification = self._swagger_api_spec.to_dict()
            specification = self._enable_api_auth(specification)
        return specification

    def _init_pweb_schema(self):
        self._add_component_schema(SwaggerCommon.MESSAGE_RESPONSE, PWebSwaggerSchema.pweb_api_message_response_schema())
        self._add_component_schema(SwaggerCommon.ERROR_DETAILS_RESPONSE, PWebSwaggerSchema.pweb_api_error_response_schema())

    def _add_component_schema(self, key: str, data):
        if key not in self._swagger_api_spec.components.schemas:
            self._swagger_api_spec.components.schema(key, schema=data)

    def _entry_spec(self, definition: SwaggerParamDef):
        if definition.url:
            self._swagger_api_spec.path(
                path=definition.url,
                parameters=self._get_query_and_url_parameters(definition),
                operations=self._get_operations(definition)
            )

    def _get_query_and_url_parameters(self, definition: SwaggerParamDef):
        parameters = []
        self._process_get_request_parameters(definition.query_params, parameters, PWebSwaggerSchema.IN_QUERY)
        self._process_get_request_parameters(definition.url_params, parameters, PWebSwaggerSchema.IN_PATH)

        if parameters:
            return parameters
        return None

    def _process_get_request_parameters(self, params, parameters, place):
        if params:
            for query in params:
                if isinstance(query, tuple) and len(query) != 0:
                    name = self._get_tuple_value(query, 0)
                    data_type = self._get_tuple_value(query, 1, SwaggerDataType.string)
                    is_required = self._get_tuple_value(query, 2, False)
                    parameters.append(PWebSwaggerSchema.get_url_param_schema(place, name, data_type, is_required))
        return parameters

    def _get_tuple_value(self, data: tuple, index: int, default=None):
        try:
            return data[index]
        except:
            return default

    def _get_operations(self, definition: SwaggerParamDef):
        operations = {}
        for method in definition.methods:
            request_body = self._process_and_add_request_schema(definition)
            responses = self._process_and_add_response_schema(definition)
            method = method.lower()
            operations[method] = {}
            if request_body:
                operations[method]["requestBody"] = request_body
            if responses:
                operations[method]["responses"] = responses
            if definition.tags:
                operations[method]["tags"] = definition.tags
        if operations:
            return operations
        return None

    def _process_and_add_request_schema(self, definition: SwaggerParamDef):
        request_schema = None
        many = False
        if definition.request_obj:
            request_schema = definition.request_obj
            if definition.def_type != SwaggerDefinitionType.FORM_DATA and definition.def_type != SwaggerDefinitionType.FILE_UPLOAD:
                request_schema = PWebSwaggerSchema.pweb_api_data_schema(definition.request_obj)

        elif definition.request_list:
            many = True
            request_schema = PWebSwaggerSchema.pweb_api_data_schema(definition.request_list, many=many)

        if request_schema:
            self._add_component_schema(definition.request_schema_key, request_schema)
            return PWebSwaggerSchema.get_request_body(definition, many=many)
        return None

    def _process_and_add_response_schema(self, definition: SwaggerParamDef):
        response_schema = None
        many = False
        if definition.response_obj:
            response_schema = definition.response_obj
        elif definition.response_list:
            many = True
            response_schema = definition.response_list

        if definition.def_type == SwaggerDefinitionType.PAGINATION:
            response_schema = PWebSwaggerSchema.pweb_api_paginate_response_schema(response_schema)
        else:
            response_schema = PWebSwaggerSchema.pweb_api_response_data_schema(response_schema, many=many)

        if response_schema:
            self._add_component_schema(definition.response_schema_key, response_schema)
            return PWebSwaggerSchema.get_response_body(definition, many=many)
        return None

    def _enable_api_auth(self, definition: dict):
        if PWebFRConfig.ENABLED_JWT_AUTH:
            if definition and isinstance(definition, dict):
                definition['security'] = [{"bearerAuth": []}]
                if "components" not in definition:
                    definition["components"] = {}
                definition["components"]["securitySchemes"] = {
                    "bearerAuth": {
                        "type": "http",
                        "scheme": "bearer",
                        "bearerFormat": "JWT"
                    }
                }
        return definition



from pweb_form_rest.crud.pweb_request_data import RequestData
from pweb_form_rest.form.pweb_form_field import FormField
from pweb_form_rest.form.pweb_process_form_field import ProcessFormFiled


class FormDefinition:
    is_validation_error: bool = False
    process_form_filed = ProcessFormFiled()
    request_data: RequestData = RequestData()
    _record_form_fields: list = []
    _record_check_uncheck_fields: list = []

    def init(self, declared_fields: dict = None):
        self.is_validation_error = False
        self._process_form_field(declared_fields)

    def _process_form_field(self, declared_fields: dict, existing_form_definition=None):
        for field_name in declared_fields:
            declared_field_definition = declared_fields[field_name]
            if declared_field_definition and not declared_field_definition.dump_only:
                if field_name not in self._record_form_fields:
                    self._record_form_fields.append(field_name)
                form_field: FormField = self._convert_declared_field_to_form_field(declared_field_definition, existing_form_definition)
                if form_field.inputType == "radio" or form_field.inputType == "checkbox":
                    self._record_check_uncheck_fields.append(field_name)
                setattr(self, form_field.name, form_field)

    def _convert_declared_field_to_form_field(self, declared_field_definition, existing_form_definition=None):
        form_field: FormField = FormField()
        if existing_form_definition and hasattr(existing_form_definition, declared_field_definition.name):
            form_field = getattr(existing_form_definition, declared_field_definition.name)
        form_field = self.process_form_filed.start(declared_field_definition, form_field=form_field)
        return form_field

    def get_form_field(self, field_name):
        if hasattr(self, field_name):
            return getattr(self, field_name)
        return None

    def set_select_option(self, field_name, select_options: dict):
        form_field = self.get_form_field(field_name=field_name)
        if not form_field or not select_options or not isinstance(select_options, dict):
            return False
        form_field.selectOptions = select_options
        return True

    def process_and_set_option(self, field_name, options: list, key_name: str, value_name: str):
        select_options = {}
        for option in options:
            if key_name in option and value_name in option:
                select_options[option[key_name]] = option[value_name]
        return self.set_select_option(field_name=field_name, select_options=select_options)

    def set_field_errors(self, errors: dict):
        self.is_validation_error = True
        for field_name in errors:
            if hasattr(self, field_name):
                form_field: FormField = getattr(self, field_name)
                form_field.errorText = errors[field_name]
                form_field.isError = True

    def set_field_value(self, field_and_value: dict):
        for field_name in field_and_value:
            if hasattr(self, field_name) and field_and_value[field_name] != None:
                form_field: FormField = getattr(self, field_name)
                form_field.value = field_and_value[field_name]

    def set_dict_value(self, name_value: dict):
        if name_value and isinstance(name_value, dict):
            self.set_field_value(name_value)
            return True
        return False

    def set_value(self, field_name, value):
        if field_name and value:
            self.set_field_value({field_name: value})
            return True
        return False

    def set_model_value(self, model):
        for field_name in self._record_form_fields:
            if hasattr(self, field_name) and hasattr(model, field_name):
                value = getattr(model, field_name)
                self.set_value(field_name, value)

    def get_request_data(self, all_data=None):
        if not all_data:
            all_data = self.request_data.form_and_file_data()

        if not all_data:
            return all_data

        for field_name in list(all_data.keys()):
            if hasattr(self, field_name):
                form_field: FormField = getattr(self, field_name)
                if form_field.dataType in ["Integer", "Float", "DateTime", "Date", "EnumField", "FileField", "Boolean"] and all_data[field_name] == "":
                    all_data[field_name] = None
                if form_field.required and (all_data[field_name] == "" or all_data[field_name] is None):
                    del all_data[field_name]

        if self._record_check_uncheck_fields:
            for check_uncheck_input in self._record_check_uncheck_fields:
                if check_uncheck_input not in all_data and hasattr(self, check_uncheck_input):
                    form_field: FormField = getattr(self, check_uncheck_input)
                    all_data[check_uncheck_input] = form_field.unchecked

        return all_data



from copy import copy
from flask import flash, redirect
from marshmallow import ValidationError
from werkzeug.datastructures import FileStorage
from pweb_auth.common.pweb_auth_config import PWebAuthConfig
from pweb_form_rest import FileDataCRUD
from pweb_form_rest.common.pweb_fr_exception import FormRESTException
from pweb_form_rest.crud.pweb_crud_common import PWebCRUDCommon
from pweb_form_rest.form.pweb_form import PWebForm
from pweb_form_rest.form.pweb_form_data import PWebFormData
from pweb_form_rest.ui.pweb_ui_helper import PWebSSRUIHelper, ssr_ui_render
from pweb_orm import PWebBaseModel


class FormDataCRUD(PWebCRUDCommon):
    ssr_ui_helper: PWebSSRUIHelper = None
    file_data_crud: FileDataCRUD = None
    pweb_form_data: PWebFormData = PWebFormData()

    def __init__(self, model: PWebBaseModel, ssr_ui_helper: PWebSSRUIHelper = None):
        self.model = model
        self.ssr_ui_helper = ssr_ui_helper
        self.file_data_crud = FileDataCRUD(model=self.model)

    def render(self, view_name, params: dict = None, form: PWebForm = None):
        return ssr_ui_render(view_name=view_name, params=params, form=form, ssr_ui_helper=self.ssr_ui_helper)

    def pre_process_file_upload(self, request_data):
        processed_request_data = copy(request_data)
        requested_files = {}
        for field_name in request_data:
            if isinstance(request_data[field_name], FileStorage):
                filename = request_data[field_name].filename.lower()
                if filename and filename != "":
                    processed_request_data[field_name] = filename
                    requested_files[field_name] = request_data[field_name]
                elif filename == "":
                    del processed_request_data[field_name]
        return processed_request_data, requested_files

    def post_process_file_upload(self, requested_files, model, form: PWebForm, upload_path: str, override_names: dict = None):
        if not requested_files or not upload_path:
            return model
        return self.file_data_crud.upload_and_save_file(form_data=requested_files, model=model, request_dto=form, override_names=override_names, upload_path=upload_path)

    def handle_various_exception(self, exception, form: PWebForm):
        if exception and form:
            if isinstance(exception, ValidationError):
                self.pweb_form_data.handle_validation_exception(exception=exception, definition=form.definition)
            elif isinstance(exception, FormRESTException):
                self.pweb_form_data.handle_form_rest_exception(exception=exception, definition=form.definition)
        message = PWebAuthConfig.CHECK_VALIDATION_ERROR_SM
        if exception and hasattr(exception, "message"):
            message = exception.message
        flash(message, "error")

    def only_create(self, form: PWebForm, data: dict = None, upload_path: str = None, override_names: dict = None):
        if form.is_post_data() and form.is_valid_data(form_data=data):
            request_data = form.get_request_data(form_data=data)
            processed_request_data, requested_files = self.pre_process_file_upload(request_data=request_data)
            model = self.save(request_dto=form, data=processed_request_data)
            self.post_process_file_upload(requested_files=requested_files, model=model, form=form, upload_path=upload_path, override_names=override_names)
            return model
        return None

    def only_update(self, update_form: PWebForm, model_id: int = None, existing_model=None, data: dict = None, query=None, upload_path: str = None, override_names: dict = None):
        if update_form.is_post_data() and update_form.is_valid_data(form_data=data):
            request_data = update_form.get_request_data(form_data=data)
            processed_request_data, requested_files = self.pre_process_file_upload(request_data=request_data)
            model = self.edit(request_dto=update_form, model_id=model_id, existing_model=existing_model, data=processed_request_data, query=query)
            self.post_process_file_upload(requested_files=requested_files, model=model, form=update_form, upload_path=upload_path, override_names=override_names)
            return model
        return None

    def create(self, view_name: str, form: PWebForm, redirect_url=None, data: dict = None, params: dict = None, response_message: str = "Successfully created!", upload_path: str = None, override_names: dict = None):
        if form.is_post_data() and form.is_valid_data(form_data=data):
            model = self.only_create(form=form, data=data, upload_path=upload_path, override_names=override_names)
            flash(response_message, "success")
            if model and redirect_url:
                return redirect(redirect_url)
            if model:
                return model
        return self.render(view_name=view_name, form=form, params=params)

    def update(self, view_name: str, update_form: PWebForm, model_id: int = None, display_from: PWebForm = None, redirect_url: str = None, details_model=None, params: dict = None, response_message: str = "Successfully updated!", existing_model=None, data: dict = None, query=None, upload_path: str = None, override_names: dict = None):
        if update_form.is_post_data() and update_form.is_valid_data(form_data=data):
            model = self.only_update(update_form=update_form, model_id=model_id, existing_model=existing_model, data=data, query=query, upload_path=upload_path, override_names=override_names)
            flash(response_message, "success")
            if model and redirect_url:
                return redirect(redirect_url)
            if model:
                return model
        elif update_form.is_get_data():
            if not details_model and model_id:
                details_model = self.get_by_id(model_id=model_id, exception=False, query=query)

            if not details_model:
                flash('Invalid data', 'error')
                if redirect_url:
                    return redirect(redirect_url)

            if display_from and details_model:
                update_form.set_dict_value(display_from.dump(details_model))
                if hasattr(details_model, "id"):
                    update_form.set_value("id", details_model.id)
            else:
                update_form.set_model_value(details_model)

        if not params:
            params = {}
        params["isEdit"] = True
        return self.render(view_name=view_name, form=update_form, params=params)

    def delete(self, model_id: int, redirect_url: str, response_message: str = "Successfully deleted!", error_message: str = "Sorry unable to delete", query=None):
        is_deleted = self.soft_remove(model_id=model_id, query=query, exception=False)
        if is_deleted:
            flash(response_message, 'success')
        else:
            flash(error_message, 'error')
        return redirect(redirect_url)

    def details(self, view_name, model_id: int, redirect_url: str, display_from: PWebForm = None, params: dict = None, query=None):
        data = self.get_by_id(model_id=model_id, query=query, exception=False)
        if not data:
            flash('Invalid data', 'error')
            return redirect(redirect_url)

        if display_from:
            data = display_from.dump(data)

        if not params:
            params = {}

        params.update({"data": data})
        return self.render(view_name=view_name, params=params)

    def paginated_list(self, view_name, response_def: PWebForm = None, query=None, search_fields: list = None, sort_field=None, sort_order=None, item_per_page=None, params: dict = None):
        data_list = self.read_all(query=query, search_fields=search_fields, sort_field=sort_field, sort_order=sort_order, item_per_page=item_per_page)
        if response_def and data_list.items:
            data_list.items = response_def.dump(data_list.items, many=True)
        if not params:
            params = {}
        params.update({"table": data_list})
        return self.render(view_name=view_name, params=params)

    def list(self, view_name, response_def: PWebForm = None, query=None, search_fields: list = None, sort_field=None, sort_order=None, params: dict = None):
        data_list = self.read_all(query=query, search_fields=search_fields, sort_field=sort_field, sort_order=sort_order, enable_pagination=False)
        if response_def:
            data_list = response_def.dump(data_list.items, many=True)
        if not params:
            params = {}
        params.update({"table": data_list})
        return self.render(view_name=view_name, params=params)



import random
import string
from pweb_form_rest.common.pweb_fr_constant import SwaggerDefinitionType, HTTPContentType


class SwaggerParamDef(object):
    url: str = None
    query_params: list = None  # list of tuple [(name, data_type, required)]
    url_params: list = None  # list of tuple [(name, data_type, required)]
    request_obj = None
    request_list = None
    response_obj = None
    response_list = None

    def_type: str = SwaggerDefinitionType.NONE

    method: str = None
    tag: str = None
    methods: list = []
    tags: list = []
    description: str = ""
    response_content_type: str = HTTPContentType.APPLICATION_JSON
    request_content_type: str = HTTPContentType.APPLICATION_JSON

    pweb_message_response: bool = False
    pweb_error_details_response: bool = False

    http_response_code: int = 200

    # Only for Internal Use
    request_schema_key: str = None
    response_schema_key: str = None

    def init_schema_key(self):
        component_code = ''.join(random.choice(string.ascii_lowercase) for i in range(12))
        self.request_schema_key = "req_" + component_code
        self.response_schema_key = "res_" + component_code



class PWebFRConfig:
    # API Config
    PAGE_PARAM_NAME: str = "page"
    ITEM_PER_PAGE_PARAM_NAME: str = "per-page"
    SORT_FIELD_PARAM_NAME: str = "sort-field"
    SORT_ORDER_PARAM_NAME: str = "sort-order"
    SEARCH_FIELD_PARAM_NAME: str = "search"
    SORT_DEFAULT_ORDER_NAME: str = "desc"
    SORT_DEFAULT_FIELD_NAME: str = "id"
    TOTAL_ITEM_PER_PAGE: int = 25

    # Swagger UI Config
    ENABLE_SWAGGER_UI: bool = True
    ENABLE_SWAGGER_AUTH: bool = False
    SWAGGER_AUTH_USERNAME: str = "pweb"
    SWAGGER_AUTH_PASSWORD: str = "pweb12"
    SWAGGER_JSON_URL: str = "/pweb-swagger-json"
    SWAGGER_UI_URL: str = "/pweb-swagger-ui"
    SWAGGER_UI_ASSETS_URL: str = "/swagger-assets"
    SWAGGER_DEFAULT_TAG_NAME: str = "Common"
    SWAGGER_TITLE: str = "PWeb Swagger"
    SWAGGER_APP_VERSION: str = "1.0.0"

    ENABLED_JWT_AUTH: bool = True

    # Messages
    INVALID_REQUEST_DATA: str = "Invalid Request Data"
    VALIDATION_ERROR: str = "Validation Error!"
    UNKNOWN_ERROR: str = "Unknown Error Occurred!"

    # File Upload
    FILE_SIZE_NOT_MATCH: str = "File size is bigger than allowed"
    INVALID_FILE_EXTENSION: str = "Invalid uploaded file"
    INVALID_FILE_UPLOAD_PATH: str = "Invalid upload path"

    # Form Helper
    SSR_UI_HELPER: "PWebSSRUIHelper" = None



from pweb_form_rest.form.pweb_form_data import PWebFormData
from pweb_form_rest.form.pweb_form_definition import FormDefinition
from pweb_form_rest.schema.pweb_rest_schema import PWebRestDTO


class PWebBaseForm:
    pass


class PWebForm(PWebBaseForm, PWebRestDTO):
    definition: FormDefinition = None
    pweb_form_data: PWebFormData = PWebFormData()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.definition = FormDefinition()
        self.definition.init(self.declared_fields)

    def set_select_option(self, field_name, select_options: dict):
        return self.definition.set_select_option(field_name=field_name, select_options=select_options)

    def process_and_set_option(self, field_name, options: list, key_name: str, value_name: str):
        return self.definition.process_and_set_option(field_name=field_name, options=options, key_name=key_name, value_name=value_name)

    def is_post_data(self) -> bool:
        return self.pweb_form_data.is_post_data()

    def is_get_data(self) -> bool:
        return self.pweb_form_data.is_get_data()

    def is_valid_data_submit(self, form_data=None):
        if self.is_post_data() and self.is_valid_data(form_data=form_data):
            return True
        return False

    def is_valid_data(self, form_data=None) -> bool:
        return self.pweb_form_data.is_valid_data(form=self, form_data=form_data, definition=self.definition)

    def get_request_data(self, form_data=None):
        return self.definition.get_request_data(all_data=form_data)

    def set_value(self, field_name, value):
        return self.definition.set_value(field_name=field_name, value=value)

    def set_dict_value(self, name_value: dict):
        return self.definition.set_dict_value(name_value=name_value)

    def set_model_value(self, model):
        return self.definition.set_model_value(model=model)

    def set_field_error(self, field_name, error):
        if self.definition:
            self.definition.set_field_errors({field_name: error})



class FormField(object):
    name: str = None
    label: str = None
    value = ""
    inputType: str = None
    defaultValue = None
    dataType: str = None
    required: bool = False
    placeholder = None
    errorText: str = None
    isErrorTextOn: bool = True
    isError: bool = False
    helpText: str = None
    isWrapper: bool = True
    isLabel: bool = True
    allAttributes: dict = None

    # Select Input
    isMultiSelect: bool = False
    selectOptions: dict = None

    # Radio & Check
    radioItem: dict = None
    checked: str = "True"
    unchecked: str = "False"

    def add_attribute(self, name, value):
        if not self.allAttributes:
            self.allAttributes = {}
        if name not in self.allAttributes:
            self.allAttributes[name] = value
        elif name in self.allAttributes:
            self.allAttributes[name] = f"{self.allAttributes[name]} {value}"



from werkzeug.datastructures import FileStorage
from ppy_file_text import FileUtil
from pweb_form_rest import FileField
from pweb_form_rest.common.pweb_fr_config import PWebFRConfig
from pweb_form_rest.common.pweb_fr_exception import form_rest_exception
from pweb_form_rest.common.pweb_web_file_util import PWebWebFileUtil
from pweb_form_rest.schema.pweb_rest_schema import PWebDataDTO


class PWebUploadProcessor:

    def validate_file_size(self, file_storage: FileStorage, field: FileField):
        if field.max_size_kb and not PWebWebFileUtil.is_valid_file_size(file_storage, field.max_size_kb):
            raise form_rest_exception.error_message_exception(PWebFRConfig.FILE_SIZE_NOT_MATCH)
        return True

    def validate_file_extension(self, file_storage: FileStorage, field: FileField):
        if field.allowed_extensions and not PWebWebFileUtil.allowed_file(file_storage.filename, field.allowed_extensions):
            raise form_rest_exception.error_message_exception(PWebFRConfig.INVALID_FILE_EXTENSION)
        return True

    def upload_file(self, input_name, file_storage: FileStorage, upload_path, override_name: dict = None, override: bool = True):
        filename = file_storage.filename.lower()
        filename = PWebWebFileUtil.process_file_name(filename)
        if override_name and input_name in override_name:
            filename = override_name[input_name] + "." + PWebWebFileUtil.get_file_extension(filename)
        if filename:
            filename = filename.lower()
        FileUtil.create_directories(upload_path)
        file_upload_path = FileUtil.join_path(upload_path, filename)
        if override:
            FileUtil.delete(file_upload_path)
            file_storage.save(file_upload_path)
        return filename

    def delete_file(self, upload_path, filename):
        file_upload_path = FileUtil.join_path(upload_path, filename)
        FileUtil.delete(file_upload_path)

    def validate_and_upload_multiple(self, form_data: dict, request_dto: PWebDataDTO, upload_path, override_name: dict = None, override: bool = True):
        response_file_name_dict = {}

        for input_name in form_data:
            file_list = form_data[input_name]

            if not isinstance(file_list, list):
                file_list = [file_list]

            override_name_index = 0
            uploaded_file_names = []
            file_name = self._get_upload_file_override_name(input_name, override_name, override_name_index)
            if file_name:
                uploaded_file_names = {}

            for file in file_list:
                file_name = self._get_upload_file_override_name(input_name, override_name, override_name_index)
                override_name_index += 1
                _override_name = None

                if file_name:
                    _override_name = {input_name: file_name}
                response = self.validate_and_upload(
                    files={input_name: file},
                    request_dto=request_dto,
                    upload_path=upload_path,
                    override_name=_override_name,
                    override=override)

                if isinstance(uploaded_file_names, list):
                    uploaded_file_names.append(response[input_name])
                elif isinstance(uploaded_file_names, dict) and file_name:
                    uploaded_file_names[file_name] = response[input_name]

            response_file_name_dict.update({input_name: uploaded_file_names})
        return response_file_name_dict

    def validate_and_upload(self, files: dict, request_dto: PWebDataDTO, upload_path, override_name: dict = None, override: bool = True):
        errors = {}
        if not upload_path:
            raise form_rest_exception.error_details_exception(PWebFRConfig.INVALID_FILE_UPLOAD_PATH, errors)

        for input_name in files:
            try:
                field: FileField = self._get_file_input(input_name, request_dto)
                if field and isinstance(files[input_name], FileStorage):
                    file_storage: FileStorage = files[input_name]
                    self.validate_file_size(file_storage, field)
                    self.validate_file_extension(file_storage, field)
                    files[input_name] = self.upload_file(input_name, file_storage, upload_path, override_name, override)
            except Exception as e:
                errors[input_name] = str(e)
        if errors:
            raise form_rest_exception.error_details_exception(PWebFRConfig.VALIDATION_ERROR, errors)
        return files

    def get_file_override_names(self, uuid, request_dto: PWebDataDTO, override_names: dict = None):
        if not override_names:
            override_names = {}
        for field_name in request_dto.fields:
            field = request_dto.fields[field_name]
            if isinstance(field, FileField) and field.name not in override_names:
                prefix = ""
                if field.save_prefix:
                    prefix = field.save_prefix + "-"
                override_names[field.name] = (prefix + uuid).lower()
        return override_names

    def delete_orphan_files(self, override_names: dict, form_data, model, upload_path):
        for name in override_names:
            if "deletedItem" in form_data and name in form_data["deletedItem"]:
                if hasattr(model, name) and (name not in form_data or not isinstance(form_data[name], FileStorage)):
                    file_path = FileUtil.join_path(upload_path, getattr(model, name))
                    setattr(model, name, None)
                    FileUtil.delete(file_path)
        return model

    def _get_upload_file_override_name(self, input_name, override_name: dict, index=0):
        if not override_name:
            return None
        file_name = None
        if input_name in override_name:
            if isinstance(override_name[input_name], list):
                file_name = self._get_file_name_from_list(index, override_name[input_name])
            else:
                file_name = override_name[input_name]
        return file_name

    def _get_file_name_from_list(self, index, list_data: list, default_name: str = "file"):
        try:
            if not list_data:
                return None
            return list_data[index]
        except IndexError:
            return default_name + str(index)

    def _get_file_input(self, name: str, request_dto: PWebDataDTO):
        if name in request_dto.fields:
            field = request_dto.fields[name]
            if isinstance(field, FileField):
                return field
        return None



from ppy_file_text import StringUtil
from pweb_form_rest.form.pweb_form_field import FormField


class ProcessFormFiled:

    def start(self, declared_field_definition, form_field: FormField):
        form_field = self._set_basic_information(declared_field_definition, form_field=form_field)
        form_field = self._set_field_default_value(declared_field_definition, form_field=form_field)
        form_field = self._process_metadata(declared_field_definition, form_field=form_field)
        form_field = self._process_enum_to_select_option(declared_field_definition, form_field=form_field)
        form_field = self._set_required_error_text(declared_field_definition, form_field=form_field)
        form_field = self._set_label(form_field=form_field)
        form_field = self._set_input_type(form_field=form_field)
        return form_field

    def _set_input_type(self, form_field: FormField):
        if not form_field.inputType and form_field.dataType:
            data_type = form_field.dataType
            if data_type == "Integer" or data_type == "Float" or data_type == "Decimal":
                form_field.inputType = "number"
            elif data_type == "Email":
                form_field.inputType = "email"
            elif data_type == "EnumField":
                form_field.inputType = "select"
            elif data_type == "Date":
                form_field.inputType = "date"
            elif data_type == "FileField":
                form_field.inputType = "file"
            else:
                form_field.inputType = "text"
        return form_field

    def _get_field_data(self, declared_field_definition, name, default=None):
        if hasattr(declared_field_definition, name):
            return getattr(declared_field_definition, name)
        return default

    def _set_required_error_text(self, declared_field_definition, form_field: FormField):
        if declared_field_definition.required and hasattr(declared_field_definition, "error_messages") and "required" in declared_field_definition.error_messages:
            form_field.errorText = declared_field_definition.error_messages["required"]
        return form_field

    def _set_basic_information(self, declared_field_definition, form_field: FormField):
        form_field.name = self._get_field_data(declared_field_definition, "name", default=form_field.required)
        form_field.required = self._get_field_data(declared_field_definition, "required", default=form_field.required)
        form_field.dataType = declared_field_definition.__class__.__name__
        return form_field

    def _set_field_default_value(self, declared_field_definition, form_field: FormField):
        default_value = self._get_field_data(declared_field_definition, "defaultValue", form_field.defaultValue)
        if not form_field.value and default_value:
            form_field.value = default_value
        return form_field

    def process_attributes(self, declared_attributes: dict, form_field: FormField, ignore: list = None):
        if not declared_attributes or not isinstance(declared_attributes, dict):
            return form_field

        if not form_field.allAttributes:
            form_field.allAttributes = {}

        for attribute_name in declared_attributes:
            if ignore and attribute_name in ignore:
                continue

            if attribute_name not in form_field.allAttributes:
                form_field.allAttributes[attribute_name] = declared_attributes[attribute_name]
            elif attribute_name in form_field.allAttributes:
                form_field.allAttributes[attribute_name] += f" {declared_attributes[attribute_name]}"

        return form_field

    def _process_metadata(self, declared_field_definition, form_field: FormField):
        metadata: dict = declared_field_definition.metadata
        for name in metadata:
            if hasattr(form_field, name):
                setattr(form_field, name, metadata[name])
            elif name == "type":
                form_field.inputType = metadata[name]
            elif name == "attributes":
                form_field = self.process_attributes(metadata[name], form_field)
        return form_field

    def _set_label(self, form_field: FormField):
        if not form_field.label and form_field.name and form_field.isLabel:
            form_field.label = StringUtil.human_readable(form_field.name)
        return form_field

    def _process_enum_to_select_option(self, declared_field_definition, form_field: FormField):
        if not form_field.selectOptions and hasattr(declared_field_definition, "enumType"):
            form_field.selectOptions = {}
            for enumItem in declared_field_definition.enumType:
                form_field.selectOptions[enumItem.name] = enumItem.value
        return form_field



import enum
from marshmallow import fields, ValidationError
import typing
from werkzeug.datastructures import FileStorage


class FileField(fields.String):
    max_size_kb: int = None
    allowed_extensions: list = None
    is_multiple: bool = False
    is_string_name: bool = False
    save_prefix: str = None

    default_error_messages = {
        "invalid": "Not a valid file."
    }

    def set_max_size_kb(self, size: int) -> "FileField":
        self.max_size_kb = size
        return self

    def set_allowed_extension(self, extension: list) -> "FileField":
        self.allowed_extensions = extension
        return self

    def allow_multiple(self) -> "FileField":
        self.is_multiple = True
        return self

    def allow_string_name(self) -> "FileField":
        self.is_string_name = True
        return self

    def set_save_prefix(self, prefix: str) -> "FileField":
        self.save_prefix = prefix
        return self

    def _deserialize(self, value, attr, data, **kwargs) -> typing.Any:
        if self.is_multiple and isinstance(value, list):
            for file in value:
                if not isinstance(file, FileStorage) and not self.is_string_name:
                    raise self.make_error("invalid")
            return value

        if not isinstance(value, FileStorage) and not self.is_string_name:
            raise self.make_error("invalid")
        return value


class CustomNestedField(fields.Nested):

    def _deserialize(self, value, attr, data, partial=None, **kwargs):
        return value


class BaseEnum(enum.Enum):

    @classmethod
    def values(cls) -> list:
        return list(map(lambda c: c.value, cls))

    @classmethod
    def keys(cls) -> list:
        return list(map(lambda c: c.name, cls))

    @classmethod
    def to_map(cls) -> dict:
        data = {}
        for enum_type in cls:
            data[enum_type.name] = enum_type.value
        return data

    @classmethod
    def value_to_key(cls, value):
        for enum_type in cls:
            if enum_type.value == value:
                return enum_type.name
        return None

    def __str__(self):
        return self.value

    def is_pf_enum(self):
        return True


def validate_enum_value(values: list, value: str, key: str, message: str = "Value should be any of "):
    if value not in values:
        message += '(' + ', '.join(values) + ')'
        raise ValidationError(message, key)


class EnumField(fields.String):
    enumType: BaseEnum

    def __init__(self, enumType, *args, **kwargs):
        self.enumType = enumType
        super(EnumField, self).__init__(*args, **kwargs)

    def _serialize(self, value, attr, obj, **kwargs):
        if isinstance(value, enum.Enum):
            return value.value
        return value

    def _deserialize(self, value, attr, data, **kwargs):
        if hasattr(self.enumType, 'is_pf_enum'):
            validate_enum_value(self.enumType.values(), data[attr], attr)
        name = self.enumType.value_to_key(data[attr])
        return self.enumType[name]



from marshmallow import Schema, fields
from marshmallow_sqlalchemy import SQLAlchemySchema


class APIBase(Schema):
    class Meta:
        ordered = True


class APIBaseResponse(APIBase):
    class Meta:
        ordered = True

    status = fields.String()
    code = fields.String()


class APIMessageResponse(APIBaseResponse):
    message = fields.String()


class APIErrorResponse(APIBaseResponse):
    error = fields.Dict(keys=fields.String(), values=fields.String())


class APIDataResponse(APIBaseResponse):
    message = fields.String()
    data = fields.Dict()


class APIDataListResponse(APIBaseResponse):
    data = fields.List(fields.Dict)


class Pagination(APIBase):
    page = fields.Integer()
    itemPerPage = fields.Integer()
    total = fields.Integer()
    totalPage = fields.Integer()


class APIPaginateResponse(APIDataListResponse):
    pagination = fields.Nested(Pagination())


class PWebDataDTO(APIBase, object):
    pass


class PWebOrmDTO(PWebDataDTO, SQLAlchemySchema):
    pass


class PWebRestDTO(PWebOrmDTO):
    class Meta:
        ordered = True

    id = fields.Integer(dump_only=True)
    created = fields.DateTime(dump_only=True)
    updated = fields.DateTime(dump_only=True)
    uuid = fields.UUID(dump_only=True)



from pweb_form_rest.crud.pweb_crud import PWebCRUD
from pweb_form_rest.crud.pweb_response_maker import ResponseMaker
from pweb_form_rest.schema.pweb_rest_schema import PWebDataDTO
from pweb_orm import PWebBaseModel


class PWebCRUDCommon:
    model: PWebBaseModel = None
    pweb_crud: PWebCRUD = PWebCRUD()
    response_maker: ResponseMaker = ResponseMaker()

    def message_or_data_response(self, model, response_dto: PWebDataDTO = None, response_message: str = None):
        if not response_dto:
            return self.response_maker.success_message(response_message)
        return self.response_maker.data_response(model, response_dto, message=response_message)

    def validate_data(self, data: dict, data_dto: PWebDataDTO):
        return self.pweb_crud.validate_data(data=data, data_dto=data_dto)

    def load_model_from_dict(self, data: dict, data_dto: PWebDataDTO, instance=None):
        return self.pweb_crud.load_model_from_dict(data=data, data_dto=data_dto, instance=instance)

    def get_json_data(self, data_dto: PWebDataDTO, is_validate=True, load_only=False):
        return self.pweb_crud.get_json_data(data_dto=data_dto, is_validate=is_validate, load_only=load_only)

    def get_form_data(self, data_dto: PWebDataDTO, is_validate=True, load_only=False, is_populate_model=False):
        return self.pweb_crud.get_form_data(data_dto=data_dto, is_validate=is_validate, is_populate_model=is_populate_model, load_only=load_only)

    def check_unique(self, field: str, value, model_id=None, exception: bool = True, message: str = "Already used", query=None):
        self.pweb_crud.check_unique(self.model, field=field, value=value, exception=exception, message=message, query=query, model_id=model_id)

    def get_by_id(self, model_id, exception=True, message: str = "Entry Not Found!", query=None):
        return self.pweb_crud.get_by_id(self.model, id=model_id, exception=exception, message=message, query=query)

    def read_all(self, query=None, search_fields: list = None, sort_field=None, sort_order=None, item_per_page=None, enable_pagination=True):
        return self.pweb_crud.list(model=self.model, query=query, search_fields=search_fields, sort_field=sort_field, sort_order=sort_order, item_per_page=item_per_page, enable_pagination=enable_pagination)

    def save(self, data: dict, request_dto: PWebDataDTO, existing_model=None):
        model = self.pweb_crud.populate_model(data, request_dto, instance=existing_model)
        model.save()
        return model

    def edit(self, model_id, data: dict, request_dto: PWebDataDTO, existing_model=None, query=None):
        if not existing_model:
            existing_model = self.get_by_id(model_id, query=query, exception=True)
        return self.save(data=data, request_dto=request_dto, existing_model=existing_model)

    def soft_remove(self, model_id: int, query=None, exception=True):
        existing_model = self.get_by_id(model_id, exception=exception, query=query)
        if existing_model and hasattr(existing_model, "isDeleted"):
            existing_model.isDeleted = True
            existing_model.save()
            return True
        return False

    def remove(self, model_id: int, query=None):
        existing_model = self.get_by_id(model_id, exception=True, query=query)
        existing_model.delete()
